/* Karotz "movies" application Copyright (c) 2012 Douglas Henke
   dhenke@mythopoeic.org; free software under modified BSD license.
   See COPYING for details. */

/* Author's notes:
   o- Yes, I know about tinyxmldom.js. The results page Google sends back
      is HTML (with unquoted attributes and unclosed tags), not XML.

   o- Yes, I know about htmlparser.js. I couldn't get it to work with Karotz.

   o- All of the parsing code is fairly brittle, and likely to break if
      Google makes substantial changes to the movies results page.

   o- It would be nice if there were a published API I could use. (If it's
      limited to 100 searches per day, or costs money, or means every
      individual end user has to generate an API key, that's a non-starter.)
*/

/* To log via a web service, uncomment the following and change to the
   URL for your log app. Do not define logurl for production releases. */
//logurl="http://myhost.example.org/~username/karotz-log/";

/* To connect to a Karotz device on the local network, e.g. for debugging,
   you can set host to a DNS name or dotted quad. For production releases,
   leave set to 'localhost'. */
host="localhost";

/*** hlog() -- emit diagnostic log information

If the global variable logurl is defined, tries to log the argument by
appending ?log= and the argument text to the URL, then doing an HTTP GET.
If logurl is undefined, or if HTTP logging doesn't work for some other reason,
uses the built-in log() method.
***/
function hlog(text) {
   if(typeof(logurl) == 'undefined') { log(text); return; }
   try {
      http.get(logurl+"?log="+escape(text));
   } catch(e) { log(text); }
}

/*** buttonEvent() -- callback to handle button presses

Callback to react when the button is pressed. The argument is a string
containing the event type.

On a single press, skip the current TTS phrase. On any other kind of press,
assume the user just wants out; halt all speech and exit.
***/
function buttonEvent(event) {
   if(event == "SIMPLE") { karotz.tts.stop(); return true; }
   spk = null;
   karotz.tts.stop();
   exit();
}

/*** ttsEvent() -- handle text-to-speech evetns

Callback for the text-to-speech engine. The argument is a string containing
an event type.

ERROR events cause an immediate exit. OK events are ignored. For other event
types, we remove the first (zeroth) element of spk and start saying it (or,
if spk is empty, just exit).

To use this, spk.push() the phrases you want to say, then call
ttsEvent("GO") to start talking. Your app will exit when the speech is done.
***/
var spk = new Array();
function ttsEvent(event) {
   if(event == "ERROR") exit();
   if(event == "OK") return true;

   /* for CANCELLED, TERMINATED or unrecognized event types: */
   if(spk == null || spk.length == 0) exit(); /* exit if nothing more to say */
   karotz.tts.start(spk.splice(0,1), "en", ttsEvent); /* say the next thing */
   return true;
}

/*** xmlUnescape() -- turn escapes like &#39; into characters

This is a very limited XML text unescaper. It only knows about sequences
like &#nnn; where nnn is a decimal character code.

The argument is an escaped string. Returns an unescaped version.
***/
function xmlUnescape(text) {
   var prev = 0, str = "", regexp = /&#([^;]+);/g;

   regexp.lastIndex = 0;
   while((rtn = regexp.exec(text)) != null) {
      if(prev != rtn.index) str += text.substring(prev, rtn.index);
      str += String.fromCharCode(rtn[1]);
      prev = rtn.index+rtn[0].length;
   }
   if(prev != text.length) str += text.substr(prev);
   return str;
}

/*** getelem() -- get portion of text string containing single HTML element

Given a string which starts with an HTML element, returns the substring
containing only that element. Example:

   var s="<div class=foo><div class=bar>text</div></div></div>other stuff";
   log(getelem(s));

will log "<div class=foo><div class=bar>text</div></div>". This is somewhat
specific to the HTML returned by Google on movie searches, and comes nowhere
near handling all the special cases in HTML in general. In particular, it
assumes element names are all lower-case alpha, and that all elements save
img are closed. It doesn't understand <tag /> either.
***/
function getelem(text) {
   var rtn, ename, lvl = 0, regexp = /<(\/?)([a-z]+)[^>]*>/g; 

   regexp.lastIndex = 0;
   while((rtn = regexp.exec(text)) != null) {
      if(lvl == 0) {
         if(rtn[1] != "" || rtn.index != 0)
            throw "getelem on something not an element";
         ename = rtn[2];
      }
      if(rtn[2] == "img") continue; // img tags are unclosed
      if(rtn[1] == "") lvl++;
      else lvl--;
      if(lvl == 0) {
         if(rtn[2] != ename) throw "getelem on element with wrong closer";
         return text.substring(0, rtn.index + rtn[0].length);
      }
   }
   throw "getelem on unclosed element:\n" + text;
}

/*** textof() -- get text portion of HTML element

Given a string containing an HTML element, returns a string containing the
enclosed text.
***/
function textof(elem) {
   var rtn, prev = 0, str = "", regexp = /<[^>]*>/g; 
   
   regexp.lastIndex = 0;
   while((rtn = regexp.exec(elem)) != null) {
      if(prev != rtn.index) str += elem.substring(prev, rtn.index);
      prev = rtn.index+rtn[0].length;
   }
   return str;
}

/*** parsetimes() -- turn Google times element into list of 24-hour times

The argument is a string containing the times element from the Google movie
results page. Returns a list of integers representing showtimes in 24-hour
format. (An hueristic is used to disambiguate the 12-hour times returned
by Google in cases where they're not explicitly marked as AM or PM.)

Example: 30, 1200, 1500, 2400 means the movie is showing at 12:30 AM,
noon, 3:00PM and midnight.

Returns null if the times can't be found or don't make sense.
***/
function parsetimes(text) {
   var lst, elem, ampm, jumps, jump_pos;

   text = text.replace(/&nbsp/g, ''); // remove nonbreaking spaces
   text = text.replace(/&#[0-9]+;/g, ''); // remove HTML entities
   text = text.replace(/^\s+/, ''); // remove leading whitespace
   text = text.replace(/\s+$/, ''); // remove trailing whitespace

   /* replace internal runs of contiguous whitespace with a single
      seperator character "~" */
   text = text.replace(/\s+/g, '~');

   /* Turn the string of times into a list. */
   lst = text.split("~");
   ampm = new Array();
   for(elem in lst) {
      /* If the element is marked explicitly as AM or PM, remember it: */
      if(lst[elem].match(/pm$/i) != null) ampm[elem] = 2;
      else if(lst[elem].match(/am$/i) != null) ampm[elem] = 1;
      else ampm[elem] = 0;

      /* Turn time into a number from 100 to 1259 inclusive. Check that
         we got something reasonable: */
      lst[elem]=parseInt(lst[elem].replace(/^(\d\d?):(\d\d).*/, "$1$2"));
      if(lst[elem] == NaN || lst[elem] < 100 || lst[elem] > 1259) return null;

      /* Normalize times 1200 to 1259 to 0000-0059: */
      if(lst[elem] >= 1200) lst[elem] -= 1200;
      if(ampm[elem] == 2) lst[elem] += 1200;
   }

   /* There should be zero or one places in the list where the time jumps
      backward. Count the number of such places: */
   jumps = 0;
   for(elem=0; elem < lst.length-1; elem++) {
     if(lst[elem] >= lst[elem+1]) { jumps++; jump_pos = elem; }
   }
   if(jumps > 1) return null;

   for(elem in lst) {
      if(ampm[elem] != 0) continue; /* AM/PM already known; no action needed */
      /* If no jumps, all times are PM. If there is a jump, all times after
         it are PM and all before, AM. */
      if(jumps == 0 || elem > jump_pos) lst[elem] += 1200;
   }
   return lst;
}

/*** divclass() -- get array of divs with a certain class

Given an HTML string, returns an array where each array element is a
string containing an HTML div with a given class name.

The first argument is the HTML string. The second, the class name.

NOTE: assumes a div has no attributes other than class, and that the
class name is quoted.

NOTE: nested divs with the same class name will probably not do what
you want.
***/
function divclass(text, name) {
   var regexp = new RegExp("<div[^>]*class="+name+">", "g");
   var div, rtn = new Array();
 
   while((div = regexp.exec(text)) != null) {
      rtn.push(getelem(text.substring(div.index)));
   }
   return rtn;
}

/*** mkurl() -- create the search URL given criteria

Given the search criteria, returns a string containing the full search URL.
The arguments are:
   near -- string containing a Zip/Post code or city name
   theater -- name of the theater
   day -- number of days after today to search (or zero for today)
***/
function mkurl(near, theater, day) {
   return "http://google.com/movies?" +
    "near=" + escape(near) +
    "&q=" + escape(theater) + ((day != 0) ? "&date=" + day : "");
}

/*** dayname() -- get the name of a day some number of days from now

Given an integer argument, returns the name of the weekday that many days
from (local) now. Example: If it's Friday now, dayname(3) is "Monday".
***/
function dayname(day) {
   return [ "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
    "Friday", "Saturday"] [ (new Date().getDay()+day) % 7 ];
}

/*** times_ampm() -- turn list of times into speakable string (12-hour clock)

Given a list of integers representing times of day on a 24-hour clock, returns
a string which can be passed to the text-to-speech function. Times are read
according to a twelve-hour clock. Since morning movies are the exception,
times before noon are explicitly called out am "AM" -- all other times are
assumed to be PM.
***/
function times_ampm(lst) {
   var hr, min, elem, rtn='';

   for(elem in lst) {
      hr = Math.floor(lst[elem] / 100);
      min = lst[elem] % 100;
      if(min == 0) {
         if(hr == 0 || hr == 24) { rtn += 'midnight. '; continue; }
         if(hr == 12) { rtn += 'noon. '; continue; }
      }
      if(hr == 0) hr = 12;
      else if(hr > 12) hr -= 12;
      if(min == 0) rtn += hr + ' o clock';
      else if(min < 10) rtn += hr + ' o ' + min;
      else rtn += hr + ' ' + min;
      if(lst[elem] < 1200) rtn += ' ay em';
      rtn += '. ';
   }
   return rtn;
}

/*** times_24() -- turn list of times into speakable string (24-hour clock)

Given a list of integers representing times of day on a 24-hour clock, returns
a string which can be passed to the text-to-speech function. Times are read
in military fashion, according to a 24-hour clock. Example: 12:30AM is the
number 30, and is spoken as "zero thirty". 3:30AM is 330 is "oh three
thirty". 10:45PM is 2245 is "twenty two forty five". 8:00AM is 800 is
"oh eight hundred".
***/
function times_24(lst) {
   var hr, min, elem, rtn='';

   for(elem in lst) {
      hr = Math.floor(lst[elem] / 100);
      min = lst[elem] % 100;
      if(hr == 0) rtn += 'zero ';
      else if(hr < 10) rtn += 'oh ' + hr + ' ';
      else rtn += hr + ' ';
      if(min == 0) rtn += 'hundred. ';
      else rtn += min + '. ';
   }
   return rtn;
}

/*** timefilter() -- given a list of showtimes, delete those we can't attend

This function takes a list of showtimes and removes those which are not
interesting because they're too early, too late or not far enough in the
future.

Arguments:
   lst -- list of times (integers representing local times in 24-hour format)
   notbefore -- remove any time earlier than this
   notafter -- remove any time later than this
   travel (optional) -- remove any time less than this many minutes in the
      future

Returns the filtered list (which may be empty).
***/
function timefilter(lst, notbefore, notafter, travel) {
   var lim;

   notbefore = notbefore.replace(/^(\d\d):(\d\d)$/, "$1$2");
   if(notbefore.match(/^\d{4}$/) == null) notbefore = null;
   notafter = notafter.replace(/^(\d\d):(\d\d)$/, "$1$2");
   if(notafter.match(/^\d{4}$/) == null) notafter = null;

   if(typeof(travel) == 'undefined') travel = null;
   else if(travel.match(/^\d+$/) == null) travel = null;
   else if((travel = parseInt(travel)) == NaN) travel = null;

   if(travel != null) {
      lim = new Date();
      lim = lim.getHours() * 100 + lim.getMinutes() + travel;
   } 
   
   for(elem = lst.length-1; elem >= 0; elem--) {
      if(notbefore != null && lst[elem] < notbefore) lst.splice(elem, 1);
      else if(notafter != null && lst[elem] > notafter) lst.splice(elem, 1);
      else if(travel != null && lim > lst[elem]) lst.splice(elem, 1);
   }
   return lst;
}

/*** connected() -- do all the work once we're connected to Karotz

This is effectively the main entry point for the program, called once
we're connected to the Karotz executive.

The argument is the second argument passed to karotz.start(), and it
is ignored.
***/
function connected(data) {
   var lst, elem, name, times, timelst;

   try { // assume we're running on the real device
      var day=params[instanceName].day;
      var near = params[instanceName].near;
      var theater = params[instanceName].theater;
      var what=params[instanceName].what;
      var notbefore=params[instanceName].notbefore;
      var notafter=params[instanceName].notafter;
      var travel=params[instanceName].travel;
   } catch(e) { // for local testing
      var day = "1";
      var near = '77083';
      var theater = 'Cinemark Memorial';
      var what='times';
      var notbefore='00:00';
      var notafter='21:00';
      var travel='30';
   }

   day = parseInt(day);
   
   karotz.button.addListener(buttonEvent); // obey the button

   var url = mkurl(near, theater, day);
   var html = http.get(url); // fetch the results page

   /* make sure we found exactly one theater: */
   if((lst = divclass(html, "theater")) == null || lst.length < 1) {
      hlog("theater not found");
      if(day == 0)
         spk.push("I didn't find that theater. Please check your settings.");
      else {
         /* There are two possible reasons we didn't find anything. One is
            that they specified a bogus search string. The other is that
            the search string is OK, but there are no listings yet for the
            day they specified. Search again with day=0 to distinguish the
            two cases. */
         url = mkurl(near, theater, 0);
         html = http.get(url);
         if((lst = divclass(html, "theater")) == null || lst.length < 1)
            spk.push("I didn't find that theater. Please check your settings.");
         else
            spk.push("I found your theater, but no listings for "+
             dayname(day)+". "+"You may want to try an earlier day.");
      }
      ttsEvent("GO");
      return;
   }
   if(lst.length > 1) {
      hlog(lst.length+" theaters found");
      spk.push("I found "+lst.length+" different theaters. Please use a "+
       "more specific search.");
      ttsEvent("GO");
      return;
   }

   /* speak the movie listings: */
   if((lst = divclass(html, "movie")) == null || lst.length < 1) {
      spk.push("no movies found"); ttsEvent("GO"); return;
   }
   for(elem in lst) { /* for each movie... */
      if((name = divclass(lst[elem], "name")) == null) {
         hlog("movie "+elem+" has no name");
         continue;
      }
      if(name.length != 1) {
         hlog("movie "+elem+" has multiple names");
         continue;
      }
      if((times = divclass(lst[elem], "times")) == null || times.length != 1 ||
        (timelst = parsetimes(textof(times[0]))) == null) {
         hlog("movie "+elem+" has incomprehensible showtimes");
         continue;
      }

      /* See if this movie has any showtimes we can use... */
      if(day == 0)
         timelst = timefilter(timelst, notbefore, notafter, travel);
      else
         timelst = timefilter(timelst, notbefore, notafter);
      if(timelst.length == 0) continue;

      /* If this is the first movie we found, add an intro: */
      if(spk.length == 0) {
         if(day == 0) spk.push("now playing");
         else spk.push("movies for "+dayname(day));
      }

      /* Say the title, and maybe the times too: */
      spk.push(xmlUnescape(textof(name[0])));
      if(what == 'times') spk.push(times_ampm(timelst));
      else if(what == 'times24') spk.push(times_24(timelst));
   }
   if(spk.length == 0) spk.push("Sorry, there's nothing showing.");
   ttsEvent("GO");
}

function launch(ip, port, callback, data) {
   try {
      karotz.connect(ip, port);
      karotz.start(callback, data);
   } catch(err) {
      hlog(err);
      exit();
   }
}
launch(host, 9123, connected, {});
